@@ -31,7 +31,7 @@
  *    - neglible memory + cpu overhead
  *      - will need to construct them at some point
  *    - we can actually implement get_mount_for_mount_path()
- *      correctly
+ *      correctly even when no volume monitor is constructed
  *
  *  - implement support for GMountOperation
  *    - not implemented in the HAL volume monitor and that's all
@@ -58,6 +58,7 @@
 G_LOCK_DEFINE_STATIC(proxy_vm);
 
 static DBusConnection *the_session_bus = NULL;
+static gboolean the_session_bus_is_integrated = FALSE;
 static GHashTable *the_volume_monitors = NULL;
 
 struct _GProxyVolumeMonitor {
@@ -81,6 +82,30 @@ static DBusHandlerResult filter_function (DBusConnection *connection, DBusMessag
 
 static void signal_emit_in_idle (gpointer object, const char *signal_name, gpointer other_object);
 
+static gboolean is_supported (GProxyVolumeMonitorClass *klass);
+
+/* The is_supported API is kinda lame and doesn't pass in the class,
+   so we work around this with this hack */
+typedef gboolean (*is_supported_func) (void);
+
+static GProxyVolumeMonitorClass *is_supported_classes[10] = { NULL };
+static gboolean is_supported_0 (void) { return is_supported (is_supported_classes[0]); };
+static gboolean is_supported_1 (void) { return is_supported (is_supported_classes[1]); };
+static gboolean is_supported_2 (void) { return is_supported (is_supported_classes[2]); };
+static gboolean is_supported_3 (void) { return is_supported (is_supported_classes[3]); };
+static gboolean is_supported_4 (void) { return is_supported (is_supported_classes[4]); };
+static gboolean is_supported_5 (void) { return is_supported (is_supported_classes[5]); };
+static gboolean is_supported_6 (void) { return is_supported (is_supported_classes[6]); };
+static gboolean is_supported_7 (void) { return is_supported (is_supported_classes[7]); };
+static gboolean is_supported_8 (void) { return is_supported (is_supported_classes[8]); };
+static gboolean is_supported_9 (void) { return is_supported (is_supported_classes[9]); };
+static is_supported_func is_supported_funcs[] = {
+  is_supported_0, is_supported_1, is_supported_2, is_supported_3,
+  is_supported_4, is_supported_5, is_supported_6, is_supported_7,
+  is_supported_8, is_supported_9,
+  NULL
+};
+
 static char *
 get_match_rule (GProxyVolumeMonitor *monitor)
 {
@@ -98,7 +123,11 @@ g_proxy_volume_monitor_finalize (GObject *object)
   char *match_rule;
   GObjectClass *parent_class;
 
-  parent_class = G_OBJECT_CLASS (G_OBJECT_GET_CLASS (object));
+  /* since GProxyVolumeMonitor is a non-instantiatable type we're dealing with a
+   * sub-type here. So we need to look at the grandparent sub-type to get the
+   * parent class for GProxyVolumeMonitor */
+  parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (
+                                 g_type_class_peek_parent (G_OBJECT_GET_CLASS (object))));
 
   monitor = G_PROXY_VOLUME_MONITOR (object);
 
@@ -108,6 +137,7 @@ g_proxy_volume_monitor_finalize (GObject *object)
 
   dbus_connection_remove_filter (monitor->session_bus, filter_function, monitor);
   match_rule = get_match_rule (monitor);
+  dbus_error_init (&dbus_error);
   dbus_bus_remove_match (monitor->session_bus,
                          match_rule,
                          &dbus_error);
@@ -259,42 +289,45 @@ static GMount *
 get_mount_for_mount_path (const char *mount_path,
                           GCancellable *cancellable)
 {
-  GType type;
   GMount *mount;
-  GNativeVolumeMonitorClass *klass;
+  GProxyVolumeMonitor *volume_monitor;
+  GProxyVolumeMonitorClass *klass;
+  GHashTableIter vm_hash_iter;
+  GHashTableIter vol_hash_iter;
+  GProxyMount *candidate_mount;
 
   /* This static method on GNativeVolumeMonitor is a *complete* pain
    * in the ass to deal with; we need to rework gio so it's deprecated
    * and thus never will get called.
    *
-   * For now we just implement as badly as the GUnixVolumeMonitor
-   * shipped with gio; e.g. we just return *some* object that
-   * implements GMount and we don't worry about setting the
-   * corresponding volume or ensuring it's the same reference as what
-   * one would normally get.
+   * TODO: we don't handle the case when there's no volume monitor ever
+   * constructed. See TODO at the top of this file on how to handle that.
    */
 
   mount = NULL;
-  klass = NULL;
+  G_LOCK (proxy_vm);
 
-  /* You can't hide types from me
-   *
-   * (read: _g_unix_volume_monitor_get_type() isn't exported)
-   */
-  type = g_type_from_name ("GUnixVolumeMonitor");
-  if (type == 0)
-    goto out;
+  /* First find the native volume monitor if one exists */
+  g_hash_table_iter_init (&vm_hash_iter, the_volume_monitors);
+  while (g_hash_table_iter_next (&vm_hash_iter, NULL, (gpointer) &volume_monitor)) {
+    klass = G_PROXY_VOLUME_MONITOR_CLASS (G_OBJECT_GET_CLASS (volume_monitor));
 
-  klass = G_NATIVE_VOLUME_MONITOR_CLASS (g_type_class_ref (type));
-  if (klass == NULL)
-    goto out;
-
-  if (klass->get_mount_for_mount_path != NULL)
-    mount = klass->get_mount_for_mount_path (mount_path, cancellable);
+    if (klass->is_native) {
+      /* The see if we've got a mount */
+      g_hash_table_iter_init (&vol_hash_iter, volume_monitor->mounts);
+      while (g_hash_table_iter_next (&vol_hash_iter, NULL, (gpointer) &candidate_mount)) {
+        if (g_proxy_mount_has_mount_path (candidate_mount, mount_path))
+          {
+            mount = g_object_ref (candidate_mount);
+            goto out;
+          }
+      }
+      goto out;
+    }
+  }
 
  out:
-  if (klass != NULL)
-    g_type_class_unref (klass);
+  G_UNLOCK (proxy_vm);
   return mount;
 }
 
@@ -413,13 +446,17 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
   GProxyVolumeMonitor *monitor = G_PROXY_VOLUME_MONITOR (user_data);
   DBusMessageIter iter;
   const char *id;
+  const char *the_dbus_name;
   const char *member;
   GProxyDrive *drive;
   GProxyVolume *volume;
   GProxyMount *mount;
+  GProxyVolumeMonitorClass *klass;
 
   G_LOCK (proxy_vm);
 
+  klass = G_PROXY_VOLUME_MONITOR_CLASS (G_OBJECT_GET_CLASS (monitor));
+
   member = dbus_message_get_member (message);
 
   if (dbus_message_is_signal (message, "org.gtk.Private.RemoteVolumeMonitor", "DriveChanged") ||
@@ -428,9 +465,14 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
       dbus_message_is_signal (message, "org.gtk.Private.RemoteVolumeMonitor", "DriveEjectButton")) {
 
     dbus_message_iter_init (message, &iter);
+    dbus_message_iter_get_basic (&iter, &the_dbus_name);
+    dbus_message_iter_next (&iter);
     dbus_message_iter_get_basic (&iter, &id);
     dbus_message_iter_next (&iter);
 
+    if (strcmp (the_dbus_name, klass->dbus_name) != 0)
+      goto not_for_us;
+
     if (strcmp (member, "DriveChanged") == 0)
       {
         drive = g_hash_table_lookup (monitor->drives, id);
@@ -479,9 +521,14 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
              dbus_message_is_signal (message, "org.gtk.Private.RemoteVolumeMonitor", "VolumeRemoved")) {
 
     dbus_message_iter_init (message, &iter);
+    dbus_message_iter_get_basic (&iter, &the_dbus_name);
+    dbus_message_iter_next (&iter);
     dbus_message_iter_get_basic (&iter, &id);
     dbus_message_iter_next (&iter);
 
+    if (strcmp (the_dbus_name, klass->dbus_name) != 0)
+      goto not_for_us;
+
     if (strcmp (member, "VolumeChanged") == 0)
       {
         volume = g_hash_table_lookup (monitor->volumes, id);
@@ -522,9 +569,14 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
              dbus_message_is_signal (message, "org.gtk.Private.RemoteVolumeMonitor", "MountRemoved")) {
 
     dbus_message_iter_init (message, &iter);
+    dbus_message_iter_get_basic (&iter, &the_dbus_name);
+    dbus_message_iter_next (&iter);
     dbus_message_iter_get_basic (&iter, &id);
     dbus_message_iter_next (&iter);
 
+    if (strcmp (the_dbus_name, klass->dbus_name) != 0)
+      goto not_for_us;
+
     if (strcmp (member, "MountChanged") == 0)
       {
         mount = g_hash_table_lookup (monitor->mounts, id);
@@ -570,6 +622,7 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
 
   }
 
+ not_for_us:
   G_UNLOCK (proxy_vm);
 
   return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
@@ -578,6 +631,7 @@ filter_function (DBusConnection *connection, DBusMessage *message, void *user_da
 static void
 g_proxy_volume_monitor_init (GProxyVolumeMonitor *monitor)
 {
+  g_proxy_volume_monitor_setup_session_bus_connection (TRUE);
 }
 
 static void
@@ -586,19 +640,107 @@ g_proxy_volume_monitor_class_finalize (GProxyVolumeMonitorClass *klass)
   g_free (klass->dbus_name);
 }
 
+typedef struct {
+  char *dbus_name;
+  gboolean is_native;
+  int is_supported_nr;
+} ProxyClassData;
+
+static ProxyClassData *
+proxy_class_data_new (const char *dbus_name, gboolean is_native)
+{
+  ProxyClassData *data;
+  static int is_supported_nr = 0;
+  
+  data = g_new0 (ProxyClassData, 1);
+  data->dbus_name = g_strdup (dbus_name);
+  data->is_native = is_native;
+  data->is_supported_nr = is_supported_nr++;
+
+  g_assert (is_supported_funcs[data->is_supported_nr] != NULL);
+  
+  return data;
+}
+
 static void
 g_proxy_volume_monitor_class_intern_init_pre (GProxyVolumeMonitorClass *klass, gconstpointer class_data)
 {
-  klass->dbus_name = g_strdup ((char *) class_data);
+  ProxyClassData *data = (ProxyClassData *) class_data;
+  klass->dbus_name = g_strdup (data->dbus_name);
+  klass->is_native = data->is_native;
+  klass->is_supported_nr = data->is_supported_nr;
   g_proxy_volume_monitor_class_intern_init (klass);
 }
 
 static gboolean
-is_supported (void)
+is_remote_monitor_supported (const char *dbus_name)
 {
-  if (the_session_bus != NULL)
-    return TRUE;
-  return FALSE;
+  DBusMessage *message;
+  DBusMessage *reply;
+  DBusError dbus_error;
+  dbus_bool_t is_supported;
+
+  message = NULL;
+  reply = NULL;
+  is_supported = FALSE;
+
+  message = dbus_message_new_method_call (dbus_name,
+                                          "/",
+                                          "org.gtk.Private.RemoteVolumeMonitor",
+                                          "IsSupported");
+  if (message == NULL)
+    {
+      g_warning ("Cannot allocate memory for DBusMessage");
+      goto fail;
+    }
+  dbus_error_init (&dbus_error);
+  reply = dbus_connection_send_with_reply_and_block (the_session_bus,
+                                                     message,
+                                                     -1,
+                                                     &dbus_error);
+  if (dbus_error_is_set (&dbus_error))
+    {
+      g_warning ("invoking IsSupported() failed for remote volume monitor with dbus name %s: %s: %s",
+                 dbus_name,
+                 dbus_error.name,
+                 dbus_error.message);
+      dbus_error_free (&dbus_error);
+      goto fail;
+    }
+
+  if (!dbus_message_get_args (reply, &dbus_error,
+                              DBUS_TYPE_BOOLEAN, &is_supported,
+                              DBUS_TYPE_INVALID))
+    {
+      g_warning ("Error parsing args in reply for IsSupported(): %s: %s", dbus_error.name, dbus_error.message);
+      dbus_error_free (&dbus_error);
+      goto fail;
+    }
+
+  if (!is_supported)
+    g_warning ("remote volume monitor with dbus name %s is not supported", dbus_name);
+
+ fail:
+  if (message != NULL)
+    dbus_message_unref (message);
+  if (reply != NULL)
+    dbus_message_unref (reply);
+  return is_supported;
+}
+
+static gboolean
+is_supported (GProxyVolumeMonitorClass *klass)
+{
+  gboolean res;
+
+  G_LOCK (proxy_vm);
+  res = g_proxy_volume_monitor_setup_session_bus_connection (FALSE);
+  G_UNLOCK (proxy_vm);
+  
+  if (res)
+    res = is_remote_monitor_supported (klass->dbus_name);
+
+  return res;
 }
 
 static GVolume *
@@ -657,6 +799,7 @@ g_proxy_volume_monitor_class_init (GProxyVolumeMonitorClass *klass)
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GVolumeMonitorClass *monitor_class = G_VOLUME_MONITOR_CLASS (klass);
   GNativeVolumeMonitorClass *native_class = G_NATIVE_VOLUME_MONITOR_CLASS (klass);
+  int i;
 
   gobject_class->constructor = g_proxy_volume_monitor_constructor;
   gobject_class->finalize = g_proxy_volume_monitor_finalize;
@@ -667,7 +810,10 @@ g_proxy_volume_monitor_class_init (GProxyVolumeMonitorClass *klass)
   monitor_class->get_volume_for_uuid = get_volume_for_uuid;
   monitor_class->get_mount_for_uuid = get_mount_for_uuid;
   monitor_class->adopt_orphan_mount = adopt_orphan_mount;
-  monitor_class->is_supported = is_supported;
+
+  i = klass->is_supported_nr;
+  is_supported_classes[i] = klass;
+  monitor_class->is_supported = is_supported_funcs[i];
 
   native_class->get_mount_for_mount_path = get_mount_for_mount_path;
 }
@@ -836,70 +982,6 @@ _get_identifiers (DBusMessageIter *iter)
   return hash_table;
 }
 
-GIcon *
-_g_icon_new_from_serialized_data (const char *gicon_data)
-{
-  char **tokens;
-  GIcon *icon;
-  guint num_tokens;
-
-  g_return_val_if_fail (gicon_data != NULL, NULL);
-
-  icon = NULL;
-
-  tokens = g_strsplit (gicon_data, " ", 0);
-  num_tokens = g_strv_length (tokens);
-
-  if (num_tokens < 2)
-    {
-      g_warning ("malformed GIcon data \"%s\"", gicon_data);
-      goto out;
-    }
-
-  if (strcmp (tokens[0], "GFileIcon") == 0)
-    {
-      GFile *file;
-      char *unescaped_uri;
-
-      if (num_tokens != 2)
-        {
-          g_warning ("malformed GFileIcon gicon_data \"%s\"", gicon_data);
-          goto out;
-        }
-
-      unescaped_uri = g_uri_unescape_string (tokens[1], NULL);
-      file = g_file_new_for_uri (unescaped_uri);
-      icon = g_file_icon_new (file);
-      g_object_unref (file);
-      g_free (unescaped_uri);
-    }
-  else if (strcmp (tokens[0], "GThemedIcon") == 0)
-    {
-      int n;
-
-      icon = NULL;
-      for (n = 1; n < num_tokens; n++)
-        {
-          char *unescaped_name;
-
-          unescaped_name = g_uri_unescape_string (tokens[n], NULL);
-          if (icon == NULL)
-            icon = g_themed_icon_new (unescaped_name);
-          else
-            g_themed_icon_append_name (G_THEMED_ICON (icon), unescaped_name);
-          g_free (unescaped_name);
-        }
-    }
-  else
-    {
-      g_warning ("cannot parse gicon_data \"%s\"; please add support", gicon_data);
-    }
-
- out:
-  g_strfreev (tokens);
-  return icon;
-}
-
 DBusConnection *
 g_proxy_volume_monitor_get_dbus_connection (GProxyVolumeMonitor *volume_monitor)
 {
@@ -927,7 +1009,7 @@ register_volume_monitor (GTypeModule *type_module,
     (GBaseFinalizeFunc) NULL,
     (GClassInitFunc) g_proxy_volume_monitor_class_intern_init_pre,
     (GClassFinalizeFunc) g_proxy_volume_monitor_class_finalize,
-    (gconstpointer) g_strdup (dbus_name),  /* class_data (leaked!) */
+    (gconstpointer) proxy_class_data_new (dbus_name, is_native),  /* class_data (leaked!) */
     sizeof (GProxyVolumeMonitor),
     0,      /* n_preallocs */
     (GInstanceInitFunc) g_proxy_volume_monitor_init,
@@ -947,47 +1029,47 @@ register_volume_monitor (GTypeModule *type_module,
                                   priority);
 }
 
+/* Call with proxy_vm lock held */
 gboolean
-g_proxy_volume_monitor_setup_session_bus_connection (void)
+g_proxy_volume_monitor_setup_session_bus_connection (gboolean need_integration)
 {
   gboolean ret;
   DBusError dbus_error;
 
   ret = FALSE;
 
-  G_LOCK (proxy_vm);
   if (the_session_bus != NULL)
-    {
-      g_warning ("session bus connection is already up!");
-      dbus_connection_ref (the_session_bus);
-      goto out;
-    }
+    goto has_bus_already;
 
   /* This is so that system daemons can use gio
    * without spawning private dbus instances.
    * See bug 526454.
    */
   if (g_getenv ("DBUS_SESSION_BUS_ADDRESS") == NULL)
-    {
-      goto out;
-    }
+    goto out;
 
   dbus_error_init (&dbus_error);
   the_session_bus = dbus_bus_get_private (DBUS_BUS_SESSION, &dbus_error);
-  if (dbus_error_is_set (&dbus_error)) {
-    g_warning ("cannot connect to the session bus: %s: %s", dbus_error.name, dbus_error.message);
-    dbus_error_free (&dbus_error);
-    goto out;
-  }
-
-  _g_dbus_connection_integrate_with_main (the_session_bus);
+  if (dbus_error_is_set (&dbus_error))
+    {
+      g_warning ("cannot connect to the session bus: %s: %s", dbus_error.name, dbus_error.message);
+      dbus_error_free (&dbus_error);
+      goto out;
+    }
 
   the_volume_monitors = g_hash_table_new (g_direct_hash, g_direct_equal);
 
+ has_bus_already:
+  
+  if (need_integration && !the_session_bus_is_integrated)
+    {
+      _g_dbus_connection_integrate_with_main (the_session_bus);
+      the_session_bus_is_integrated = TRUE;
+    }
+  
   ret = TRUE;
 
  out:
-  G_UNLOCK (proxy_vm);
   return ret;
 }
 
@@ -997,8 +1079,9 @@ g_proxy_volume_monitor_teardown_session_bus_connection (void)
   G_LOCK (proxy_vm);
   if (the_session_bus != NULL)
     {
-      /* it would be nice to check that refcount==1 here */
-      _g_dbus_connection_remove_from_main (the_session_bus);
+      if (the_session_bus_is_integrated)
+        _g_dbus_connection_remove_from_main (the_session_bus);
+      the_session_bus_is_integrated = FALSE;      
       dbus_connection_close (the_session_bus);
       the_session_bus = NULL;
 
@@ -1008,62 +1091,6 @@ g_proxy_volume_monitor_teardown_session_bus_connection (void)
   G_UNLOCK (proxy_vm);
 }
 
-static gboolean
-is_remote_monitor_supported (const char *dbus_name)
-{
-  DBusMessage *message;
-  DBusMessage *reply;
-  DBusError dbus_error;
-  dbus_bool_t is_supported;
-
-  message = NULL;
-  reply = NULL;
-  is_supported = FALSE;
-
-  message = dbus_message_new_method_call (dbus_name,
-                                          "/",
-                                          "org.gtk.Private.RemoteVolumeMonitor",
-                                          "IsSupported");
-  if (message == NULL)
-    {
-      g_warning ("Cannot allocate memory for DBusMessage");
-      goto fail;
-    }
-  dbus_error_init (&dbus_error);
-  reply = dbus_connection_send_with_reply_and_block (the_session_bus,
-                                                     message,
-                                                     -1,
-                                                     &dbus_error);
-  if (dbus_error_is_set (&dbus_error))
-    {
-      g_warning ("invoking IsSupported() failed for remote volume monitor with dbus name %s: %s: %s",
-                 dbus_name,
-                 dbus_error.name,
-                 dbus_error.message);
-      dbus_error_free (&dbus_error);
-      goto fail;
-    }
-
-  if (!dbus_message_get_args (reply, &dbus_error,
-                              DBUS_TYPE_BOOLEAN, &is_supported,
-                              DBUS_TYPE_INVALID))
-    {
-      g_warning ("Error parsing args in reply for IsSupported(): %s: %s", dbus_error.name, dbus_error.message);
-      dbus_error_free (&dbus_error);
-      goto fail;
-    }
-
-  if (!is_supported)
-    g_warning ("remote volume monitor with dbus name %s is not supported", dbus_name);
-
- fail:
-  if (message != NULL)
-    dbus_message_unref (message);
-  if (reply != NULL)
-    dbus_message_unref (reply);
-  return is_supported;
-}
-
 void
 g_proxy_volume_monitor_register (GIOModule *module)
 {
@@ -1162,14 +1189,11 @@ g_proxy_volume_monitor_register (GIOModule *module)
               native_priority = 0;
             }
 
-          if (is_remote_monitor_supported (dbus_name))
-            {
-              register_volume_monitor (G_TYPE_MODULE (module),
-                                       type_name,
-                                       dbus_name,
-                                       is_native,
-                                       native_priority);
-            }
+          register_volume_monitor (G_TYPE_MODULE (module),
+                                   type_name,
+                                   dbus_name,
+                                   is_native,
+                                   native_priority);
 
         cont:
 
